#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <time.h>
#include <signal.h>
#include <linux/input.h>
#include <linux/uinput.h>

#define DEVICE_LEFTNUB  "/dev/input/event2"
#define DEVICE_RIGHTNUB "/dev/input/event3"
#define DEVICE_GAMEPAD  "/dev/input/event4"

int  handleLeftNub;
int  handleRightNub;
int  handleGamepad;
int  handleUInput;


int openDevice(const char * name)
{
    int handle = open(name, O_RDONLY | O_NONBLOCK);

    // Device could not be opened (non-existent, wrong permissions, etc...)
    if (handle < 0)
    {
        return -1;
    }

    return handle;
}


int openDevices()
{
    // Open left nub
    handleLeftNub = openDevice(DEVICE_LEFTNUB);

    if (handleLeftNub == -1)
    {
        return -1;
    }
  
    // Open right nub
    handleRightNub = openDevice(DEVICE_RIGHTNUB);

    if (handleRightNub == -1)
    {
        close(handleLeftNub);

        return -1;
    }

    // Open gpio
    handleGamepad = openDevice(DEVICE_GAMEPAD);

    if (handleGamepad == -1)
    {
        close(handleLeftNub);
        close(handleRightNub);

        return -1;
    }

    return 0;
}


void closeDevices()
{
    close(handleLeftNub);
    close(handleRightNub);
    close(handleGamepad);
}


int initUInput()
{
    int result;

    handleUInput = open("/dev/input/uinput", O_WRONLY | O_NONBLOCK);

    if(handleUInput < 0)
    {
        printf("Failed to open '/dev/input/uinput'! (%s)\n", strerror(errno));
        printf("Maybe module 'uinput' is not loaded.\n");

        return -1;
    }

    // Set capabilities
    // 3x2 axis
    ioctl(handleUInput, UI_SET_EVBIT, EV_ABS);
    ioctl(handleUInput, UI_SET_ABSBIT, ABS_X);          // Left X
    ioctl(handleUInput, UI_SET_ABSBIT, ABS_Y);          // Left Y
    ioctl(handleUInput, UI_SET_ABSBIT, ABS_RX);         // Right X
    ioctl(handleUInput, UI_SET_ABSBIT, ABS_RY);         // Right Y
    ioctl(handleUInput, UI_SET_ABSBIT, ABS_HAT0X);      // Hat X
    ioctl(handleUInput, UI_SET_ABSBIT, ABS_HAT0Y);      // Hat Y

    // Buttons
    ioctl(handleUInput, UI_SET_EVBIT, EV_KEY);
    ioctl(handleUInput, UI_SET_KEYBIT, BTN_A);          // Button A
    ioctl(handleUInput, UI_SET_KEYBIT, BTN_B);          // Button B
    ioctl(handleUInput, UI_SET_KEYBIT, BTN_X);          // Button X
    ioctl(handleUInput, UI_SET_KEYBIT, BTN_Y);          // Button Y
    ioctl(handleUInput, UI_SET_KEYBIT, BTN_TL);         // Button LeftShoulder
    ioctl(handleUInput, UI_SET_KEYBIT, BTN_TR);         // Button RightShoulder
    ioctl(handleUInput, UI_SET_KEYBIT, BTN_SELECT);     // Button Select
    ioctl(handleUInput, UI_SET_KEYBIT, BTN_START);      // Button Start
    

    struct uinput_user_dev uidev;
    memset(&uidev, 0, sizeof(uidev));

    snprintf(uidev.name, UINPUT_MAX_NAME_SIZE, "Pandora Gamepad");
    uidev.id.bustype = BUS_USB;
    uidev.id.vendor  = 0x0000;
    uidev.id.product = 0x0000;
    uidev.id.version = 1;
    uidev.absmin[ABS_X] = uidev.absmin[ABS_Y] = uidev.absmin[ABS_RX] = uidev.absmin[ABS_RY] = -256;
    uidev.absmax[ABS_X] = uidev.absmax[ABS_Y] = uidev.absmax[ABS_RX] = uidev.absmax[ABS_RY] =  256;
    uidev.absmin[ABS_HAT0X] = uidev.absmin[ABS_HAT0Y] = -1;
    uidev.absmax[ABS_HAT0X] = uidev.absmax[ABS_HAT0Y] =  1;

    result = write(handleUInput, &uidev, sizeof(uidev));

    if (result != sizeof(uidev))
    {
        printf("Failed to prepare uinput device!\n");

        close(handleUInput);

        return -1;
    }

    result = ioctl(handleUInput, UI_DEV_CREATE);

    if (result < 0)
    {
        printf("Failed to create uinput device! (%s)\n", strerror(errno));

        close(handleUInput);

        return -1;
    }
    
    return 0;
}


void finiUInput()
{
    ioctl(handleUInput, UI_DEV_DESTROY);

    close(handleUInput);
}


void processEvent(int handle)
{
    struct input_event event;

    if (read(handle, &event, sizeof(event)) <= 0)
    {
        // Check, if nubs have been reconfigured
        if (errno == ENODEV)
        {
            if (handle == handleLeftNub)
            {
                int h = openDevice(DEVICE_LEFTNUB);

                if (h != -1)
                {
                    close(handle);

                    handleLeftNub = h;
                }
            }

            if (handle == handleRightNub)
            {
                int h = openDevice(DEVICE_RIGHTNUB);

                if (h != -1)
                {
                    close(handle);

                    handleRightNub = h;
                }
            }
        }

        return;
    }

    // Handle nub events
    if ((handle == handleLeftNub) || (handle == handleRightNub))
    {
        // Filter out non joystick events
        if (event.type != EV_ABS)
        {
            return;
        }

        // Map right nub to second joystick
        if (handle == handleRightNub)
        {
            if (event.code == ABS_X) event.code = ABS_RX;
            if (event.code == ABS_Y) event.code = ABS_RY;
        }
    }


    // Handle gamepad events
    if (handle == handleGamepad)
    {
        switch (event.code)
        {
            case KEY_HOME:
            {
                event.code = BTN_A;

                break;
            }

            case KEY_END:
            {
                event.code = BTN_B;

                break;
            }

            case KEY_PAGEUP:
            {
                event.code = BTN_Y;

                break;
            }

            case KEY_PAGEDOWN:
            {
                event.code = BTN_X;

                break;
            }
            
            case KEY_LEFTALT:
            {
                event.code = BTN_START;

                break;
            }

            case KEY_LEFTCTRL:
            {
                event.code = BTN_SELECT;

                break;
            }

            case KEY_RIGHTSHIFT:
            {
                event.code = BTN_TL;

                break;
            }

            case KEY_RIGHTCTRL:
            {
                event.code = BTN_TR;

                break;
            }

            case KEY_UP:
            {
                event.type = EV_ABS;
                event.code = ABS_HAT0Y;
                event.value *= -1;

                break;
            }

            case KEY_DOWN:
            {
                event.type = EV_ABS;
                event.code = ABS_HAT0Y;

                break;
            }

            case KEY_LEFT:
            {
                event.type = EV_ABS;
                event.code = ABS_HAT0X;
                event.value *= -1;

                break;
            }

            case KEY_RIGHT:
            {
                event.type = EV_ABS;
                event.code = ABS_HAT0X;

                break;
            }

            default:
            {
                return;
            }
        }
    }
    
    write(handleUInput, &event, sizeof(event));
}


void processEvents()
{
    processEvent(handleLeftNub);
    processEvent(handleRightNub);
    processEvent(handleGamepad);
}


void signalHandler(int sigNo)
{
    closeDevices();

    finiUInput();

    printf("Gamepad: Terminating!\n");

    exit(0);
}


int main(int argc, char * argv[])
{
    if ((initUInput() == -1) || (openDevices() == -1))
    {
        printf("Could not initialize gamepad device. Aborting...\n");

        return -1;
    }

    // Catch signals
    signal(SIGINT, signalHandler);
    signal(SIGTERM, signalHandler);

    while(1)
    {
        processEvents();
    }
}
